from __future__ import annotations

from dataclasses import dataclass
from typing import Optional, Protocol, Sequence, cast

from sm.evaluation.sm_metrics import ScoringFn
from sm.evaluation.utils import PrecisionRecallF1


def precision_recall_f1(
    ytrue: list[Optional[str]] | list[set[str]],
    ypreds: list[Optional[str]],
    scoring_fn: Optional[ScoringFn] = None,
):
    """Calculate precision, recall, and f1. For each example, if the true label is either None or empty, it is out-of-class (e.g., dangling, nil, negative (for binary classifcation)) and the system
    should not predict anything (if the system does, it's a false positive). If the predict label is None, it means the system predicts this is out-of-class but of course it does not get any reward for
    that except for the fact that it does not get penalized for false positive.

    Difference from sklearn.metrics.precision_recall_f1_score: this function supports customizing the scoring function to calculate approximate recall.

    Args:
        ytrue: list of true labels per example. When there are more than one correct labels per example, we treat a prediction is correct if it is
            in the set of correct labels.
        ypreds: list of predicted labels per example, sorted by their likelihood in decreasing order.
        scoring_fn: the function telling how well a prediction matches a true label. Exact matching is used by default, but HierachicalScoringFn (in SemTab)
            can be used as well to calculate approximate recall at k.
    """
    if len(ytrue) == 0:
        return PrecisionRecallF1(0.0, 0.0, 0.0)

    if not isinstance(ytrue[0], (set, list, tuple)):
        ytrue = [{y} if y is not None else set() for y in cast(list[str], ytrue)]
    else:
        ytrue = cast(list[set[str]], ytrue)

    if scoring_fn is None:
        scoring_fn = ScoringFn()

    n_correct = 0
    n_predictions = sum(int(p is not None) for p in ypreds)
    n_labels = sum(int(len(y) > 0) for y in ytrue)

    for i in range(len(ytrue)):
        yipred = ypreds[i]

        if len(ytrue[i]) > 0 and yipred is not None:
            n_correct += max(
                scoring_fn.get_match_score(yipred, yitrue) for yitrue in ytrue[i]
            )
        else:
            assert (
                len(ytrue[i]) == 0 or yipred is None
            ), "To ensure that we don't count in case where there is no label and the system do not predict anything"

    precision = n_correct / n_predictions if n_predictions > 0 else 1.0
    recall = n_correct / n_labels if n_labels > 0 else 1.0
    f1 = (
        2 * precision * recall / (precision + recall) if precision + recall > 0 else 0.0
    )

    return PrecisionRecallF1(precision, recall, f1)



import functools
import re
from typing import List, Optional


class Column:
    # conflict with cached_property
    # __slots__ = ("index", "name", "values")

    def __init__(self, index: int, name: Optional[str], values: List):
        """
        :param index: index of the column in the original table
        :param name: name of the column, None mean the column doesn't have any name (different from having empty name)
        :param values: values in each row
        """
        self.index = index
        self.name = name
        self.values = values

    @functools.cached_property
    def clean_name(self) -> Optional[str]:
        """Clean the name that may contain many unncessary spaces."""
        if self.name is None:
            return None

        return re.sub(r"\s+", " ", self.name).strip()

    @functools.cached_property
    def clean_multiline_name(self) -> Optional[str]:
        """Clean the name that may contain many unncessary spaces. However, this function keeps the newlines intact."""
        if self.name is None:
            return None

        key = None
        for k in ["##", "###", "#@#"]:
            if k not in self.name:
                key = k
                break
        assert key is not None, "Cannot find a key to replace newlines"
        s = re.sub(r"\n+", key, self.name)  # replace newlines with key
        s = re.sub(r"\s+", " ", s).strip()  # replace multiple spaces with one space
        s = re.sub(
            f" *{key} *", "\n", s
        )  # replace key with surrounded spaces with newline
        s = re.sub(
            f"\n+", "\n", s
        )  # replace multiple consecutive newlines with a single newline
        return s

    def __getitem__(self, item):
        return self.values[item]

    def __setitem__(self, key, value):
        self.values[key] = value

    def to_dict(self):
        return {"index": self.index, "name": self.name, "values": self.values}



import csv
import time
from contextlib import contextmanager
from pathlib import Path
from typing import Union
from warnings import warn

from deprecated import deprecated

warn(f"{__name__} is deprecated", DeprecationWarning, stacklevel=2)


@deprecated(reason="use timer4 package instead")
class TimerCount:
    def __init__(self, name: str, timer: "Timer"):
        self.name = name
        self.timer = timer
        self.start = time.time()

    def stop(self):
        self.timer.categories[self.name] += time.time() - self.start
        return self.timer


@deprecated(reason="use timer4 package instead")
class Timer:
    instance = None

    def __init__(self):
        self.categories = {}

    @staticmethod
    def get_instance():
        if Timer.instance is None:
            Timer.instance = Timer()
        return Timer.instance

    @contextmanager
    def watch(self, name: str = "default"):
        try:
            count = self.start(name)
            yield None
        finally:
            count.stop()

    @contextmanager
    def watch_and_report(
        self,
        msg: str,
        print_fn=None,
        precision: int = 3,
        append_to_file: Union[None, str, Path] = None,
    ):
        """Watch and report the timer results to stdout and optionally to a file

        Args:
            msg: message to print before the report
            print_fn: function to print the results (default is print to stdout)
            precision: number of decimal places to round to
            append_to_file: file to append the results to if it's not None
        """
        print_fn = print_fn or print
        start = time.time()

        try:
            yield None
        finally:
            end = time.time()
            print_fn(msg + f": {format(end - start, f'.{precision}f')} seconds")

        if append_to_file is not None:
            Path(append_to_file).parent.mkdir(parents=True, exist_ok=True)
            rows = []
            if not Path(append_to_file).exists():
                rows.append(["name", "time"])
            rows.append([msg, end - start])
            with open(append_to_file, "a") as f:
                writer = csv.writer(
                    f, delimiter=",", quoting=csv.QUOTE_MINIMAL, lineterminator="\n"
                )
                writer.writerows(rows)

    def start(self, name: str = "default"):
        if name not in self.categories:
            self.categories[name] = 0.0
        return TimerCount(name, self)

    def report(
        self,
        print_fn=None,
        precision: int = 3,
        append_to_file: Union[None, str, Path] = None,
    ):
        """Report the timer results to stdout and optionally to a file

        Args:
            print_fn: function to print the results
            precision: number of decimal places to round to
            append_to_file: file to append the results to if it's not None
        """
        print_fn = print_fn or print
        if len(self.categories) == 0:
            print_fn("--- Nothing to report ---")
            return

        print_fn("Runtime report:")
        for k, v in self.categories.items():
            print_fn(f"\t{k}: {format(v, f'.{precision}f')} seconds")

        if append_to_file is not None:
            rows = []
            if not Path(append_to_file).exists():
                rows.append(["name", "time"])
            for k, v in self.categories.items():
                rows.append([k, v])

            with open(append_to_file, "a") as f:
                writer = csv.writer(
                    f, delimiter=",", quoting=csv.QUOTE_MINIMAL, lineterminator="\n"
                )
                writer.writerows(rows)

    def get_time(self, name: str = "default"):
        return self.categories[name]



import shutil
from contextlib import contextmanager
from dataclasses import dataclass
from hashlib import md5
from operator import attrgetter
from pathlib import Path
from typing import Callable, Generator, Generic, List, Literal, Optional, TypeVar, Union
from urllib.parse import urlparse
from zipfile import Path as ZipPath
from zipfile import ZipFile

import orjson
from serde import json
from slugify import slugify

from sm.inputs.prelude import ColumnBasedTable, Context, Link
from sm.misc.funcs import batch, get_latest_path
from sm.misc.matrix import Matrix
from sm.outputs.semantic_model import SemanticModel

T = TypeVar("T", covariant=True)


@dataclass
class Example(Generic[T]):
    sms: List[SemanticModel]
    table: T


@dataclass
class FullTable:
    table: ColumnBasedTable
    context: Context
    links: Matrix[List[Link]]

    def keep_rows(self, row_index: List[int]):
        """Keep only the rows in the table that are in row_index."""
        self.links.data = [self.links.data[i] for i in row_index]
        self.table._df = None
        for col in self.table.columns:
            col.values = [col.values[i] for i in row_index]

    def to_dict(self):
        return {
            "version": 2,
            "table": self.table.to_dict(),
            "context": self.context.to_dict(),
            "links": [
                [[link.to_dict() for link in cell] for cell in row]
                for row in self.links.data
            ],
        }

    @classmethod
    def from_dict(cls, obj: dict):
        version = obj["version"]
        if not (version == "1.2" or version == "1.1" or version == 2):
            raise ValueError(f"Unknown version: {version}")

        return cls(
            table=ColumnBasedTable.from_dict(obj["table"]),
            context=Context.from_dict(obj["context"]),
            links=Matrix(
                [
                    [[Link.from_dict(link) for link in cell] for cell in row]
                    for row in obj["links"]
                ]
            ),
        )


@dataclass
class Dataset:
    location: Path

    @contextmanager
    def _open(self) -> Generator[Union[ZipPath, Path], None, None]:
        if self.is_zip_file():
            # it is a zip file
            with ZipFile(self.location, mode="r") as zf:
                root = ZipPath(zf)
                if self.description_dir(root).exists():
                    yield root

                subdirs = list(root.iterdir())
                if len(subdirs) == 1 and self.description_dir(subdirs[0]).exists():
                    yield subdirs[0]
                else:
                    raise ValueError("Invalid dataset format")
        else:
            yield self.location

    def description_dir(self, root: Union[Path, ZipPath]) -> Union[Path, ZipPath]:
        return root / "descriptions"

    def table_dir(self, root: Union[Path, ZipPath]) -> Union[Path, ZipPath]:
        return root / "tables"

    def is_zip_file(self):
        return self.location.name.endswith(".zip")

    def load(self) -> List[Example[FullTable]]:
        """Load dataset from a folder. Assuming the following structure:

        descriptions (containing semantic descriptions of tables)
        ├── <table_fs_id>
        │   ├── version.01.json
        │   ├── version.02.json
        │   └── ...
            or
        ├── <table_fs_id>.json
        ├── ...
        tables (containing list of tables, the type of table depends on )
        ├── <table_fs_id>.json[.gz|.bz2|.lz4]
        ├── ...

        We also support compressing formats such as .zip.
        descriptions
        ├── part-<num>.zip (no nested version folders)
        │   ├── <table_fs_id>.json
        |   |   ...
        tables
        ├── part-<num>.zip
        │   ├── <table_fs_id>.json
        |   |   ...

        Args:
            data_dir:

        Returns:
        """
        examples = []
        with self._open() as root:
            descdir = self.description_dir(root)
            tabledir = self.table_dir(root)

            for infile in sorted(tabledir.iterdir(), key=attrgetter("name")):
                suffixes = Path(infile.name).suffixes
                if infile.name.startswith(".") or len(suffixes) == 0:
                    continue

                if suffixes[0] == ".json":
                    example_id = infile.name[: -sum(len(x) for x in suffixes)]
                    table = FullTable.from_dict(orjson.loads(infile.read_bytes()))

                    if descdir.exists():
                        if (descdir / example_id).exists():
                            desc_file = max(
                                (descdir / example_id).iterdir(),
                                key=lambda file: int(file.name.split(".")[1]),
                            )
                            assert desc_file is not None
                        else:
                            desc_file = descdir / f"{example_id}.json"
                            assert desc_file.exists()

                        raw_sms = orjson.loads(desc_file.read_bytes())
                        sms = [SemanticModel.from_dict(sm) for sm in raw_sms]
                    else:
                        sms = []

                    examples.append(Example(sms=sms, table=table))
                elif infile.name.endswith(".zip"):
                    assert (
                        isinstance(infile, Path)
                        and isinstance(descdir, Path)
                        and not self.is_zip_file()
                    ), "Must not be a zip file"
                    part = {}
                    with ZipFile(infile, mode="r") as zf:
                        for file in zf.infolist():
                            if not file.filename.endswith(".json"):
                                continue

                            table_id = Path(file.filename).stem
                            with zf.open(file, mode="r") as f:
                                table = FullTable.from_dict(orjson.loads(f.read()))
                            part[table_id] = table

                    if descdir.exists():
                        lst = []
                        with ZipFile(descdir / infile.name, mode="r") as zf:
                            for file in zf.infolist():
                                table_id = Path(file.filename).stem
                                if table_id not in part:
                                    continue

                                assert file.filename.endswith(".json")
                                with zf.open(file, mode="r") as f:
                                    sms = [
                                        SemanticModel.from_dict(sm)
                                        for sm in orjson.loads(f.read())
                                    ]
                                    lst.append(Example(sms=sms, table=part[table_id]))
                    else:
                        lst = [Example(sms=[], table=table) for table in part.values()]
                    assert len(lst) == len(part)
                    examples.extend(lst)

            return examples

    def save(
        self,
        examples: List[Example[FullTable]],
        individual_table_compressed: Optional[Literal["gz", "bz2", "lz4"]] = None,
        batch_compressed: bool = False,
        batch_size: int = 100,
        table_fmt_indent: Literal[0, 2] = 0,
        clean_previous_data: bool = True,
    ):
        descdir = self.description_dir(self.location)
        tabledir = self.table_dir(self.location)
        assert (
            not self.is_zip_file()
            and isinstance(descdir, Path)
            and isinstance(tabledir, Path)
        )

        if descdir.exists() and clean_previous_data:
            shutil.rmtree(descdir)
        descdir.mkdir(parents=True, exist_ok=True)

        if tabledir.exists() and clean_previous_data:
            shutil.rmtree(tabledir)
        tabledir.mkdir(parents=True, exist_ok=True)

        if batch_compressed:
            for i, bexamples in enumerate(batch(batch_size, examples)):
                bexamples: List[Example[FullTable]]
                filename = f"part-{i:04d}.zip"
                with ZipFile(descdir / filename, "w") as dzf, ZipFile(
                    tabledir / filename, "w"
                ) as tzf:
                    for e in bexamples:
                        ename = get_friendly_fs_id(e.table.table.table_id) + ".json"
                        dzf.writestr(
                            ename,
                            data=orjson.dumps([sm.to_dict() for sm in e.sms]),
                        )
                        tzf.writestr(
                            ename,
                            data=orjson.dumps(e.table.to_dict()),
                        )
        else:
            for e in examples:
                filename = get_friendly_fs_id(e.table.table.table_id)
                (descdir / filename).mkdir(exist_ok=True)
                compressed_filename = (
                    filename + f".json.{individual_table_compressed}"
                    if individual_table_compressed is not None
                    else filename + ".json"
                )
                json.ser(
                    [sm.to_dict() for sm in e.sms],
                    descdir / filename / "version.01.json",
                    indent=2,
                )
                json.ser(
                    e.table, tabledir / compressed_filename, indent=table_fmt_indent
                )


def get_friendly_fs_id(id: str) -> str:
    if id.startswith("http://") or id.startswith("https://"):
        if id.find("dbpedia.org") != -1:
            return (
                slugify(
                    urlparse(id).path.replace("/resource/", "").replace("/", "_")
                ).replace("-", "_")
                + "_"
                + md5(id.encode()).hexdigest()
            )

        if id.find("wikipedia.org") != -1:
            return (
                slugify(
                    urlparse(id).path.replace("/wiki/", "").replace("/", "_")
                ).replace("-", "_")
                + "_"
                + md5(id.encode()).hexdigest()
            )

        raise NotImplementedError()
    return slugify(id.replace("/", "_"), lowercase=False).replace("-", "_")



from __future__ import annotations
from typing import Generic, Iterable, TypeVar, Sequence
from typing_extensions import Self


V1 = TypeVar("V1")
V2 = TypeVar("V2")


class Bijection(Generic[V1, V2]):
    def __init__(self):
        self.x2prime: dict[V1, V2] = {}
        self.prime2x: dict[V2, V1] = {}

    def size(self) -> int:
        return len(self.x2prime)

    def add(self, x: V1, xprime: V2) -> bool:
        """Add a invertible mapping between x and xprime. Return true if
        successful (even when an existing mapping exists), false if x or xprime is already mapped to another value.
        """
        if (x in self.x2prime) != (xprime in self.prime2x):
            return False

        if x in self.x2prime:
            return self.x2prime[x] == xprime

        self.x2prime[x] = xprime
        self.prime2x[xprime] = x

        return True

    def check_add(self, x: V1, xprime: V2) -> Self:
        """Add a invertible mapping between x and xprime. Return Self if
        successful or raise ValueError exception if x or xprime is already mapped to another value.
        """
        flag = self.add(x, xprime)
        if not flag:
            raise ValueError(
                f"Cannot add mapping {x} <-> {xprime} because it contradicts with existing mapping {self.x2prime.get(x, '∅')} -> {self.prime2x.get(xprime, '∅')}"
            )
        return self

    def get_x(self, xprime):
        return self.prime2x[xprime]

    def get_xprime(self, x):
        return self.x2prime[x]


class IntBijection(Generic[V1]):
    """Map from an id into an integer (started from 0)"""

    def __init__(self):
        self.data = []
        self.index: dict[V1, int] = {}

    @staticmethod
    def create(data: Iterable[V1] | Sequence[V1]) -> IntBijection[V1]:
        obj = IntBijection()
        for x in data:
            if x not in obj.index:
                obj.index[x] = len(obj.data)
                obj.data.append(x)
        return obj

    def insert(self, value: V1):
        if value not in self.index:
            self.index[value] = len(self.data)
            self.data.append(value)

    def get_x(self, xprime: int):
        return self.data[xprime]

    def get_xprime(self, x: V1):
        return self.index[x]

    def get_xprimes(self, lst: list[V1]):
        return [self.index[x] for x in lst]

    def __len__(self):
        return len(self.data)

    def iter_x(self):
        return self.index.keys()

    def iter_xprime(self):
        return self.data



from __future__ import annotations

from sm.namespaces.namespace import OutOfNamespace, KnowledgeGraphNamespace
from sm.namespaces.prefix_index import PrefixIndex


class WikidataNamespace(KnowledgeGraphNamespace):
    """Namespace for Wikidata entities and ontology.

    In Wikidata, everything is an entity (classes, properties, items, etc.). But they also have decicated namespaces for their ontology predicates (/prop/).
    So for semantic models, we use the namespace for their properties instead of treating them as entities.
    For example, we use p:P131 instead of wd:P131.
    """

    __slots__ = ("entity_prefix", "property_prefix")

    ENTITY_ID = (
        "Q35120"  # the root entity of Wikidata, any other class is a subclass of this
    )
    STATEMENT_URI = (
        "http://wikiba.se/ontology#Statement"  # statement to represent n-ary relations
    )
    DUMMY_CLASS_FOR_INVERSION_URI = "http://wikiba.se/ontology#DummyClassForInversion"

    @staticmethod
    def create():
        prefix2ns = {
            "p": "http://www.wikidata.org/prop/",
            "pq": "http://www.wikidata.org/prop/qualifier/",
            "pqn": "http://www.wikidata.org/prop/qualifier/value-normalized/",
            "pqv": "http://www.wikidata.org/prop/qualifier/value/",
            "pr": "http://www.wikidata.org/prop/reference/",
            "prn": "http://www.wikidata.org/prop/reference/value-normalized/",
            "prv": "http://www.wikidata.org/prop/reference/value/",
            "ps": "http://www.wikidata.org/prop/statement/",
            "psn": "http://www.wikidata.org/prop/statement/value-normalized/",
            "psv": "http://www.wikidata.org/prop/statement/value/",
            "wd": "http://www.wikidata.org/entity/",
            "wdata": "http://www.wikidata.org/wiki/Special:EntityData/",
            "wdno": "http://www.wikidata.org/prop/novalue/",
            "wdref": "http://www.wikidata.org/reference/",
            "wds": "http://www.wikidata.org/entity/statement/",
            "wdt": "http://www.wikidata.org/prop/direct/",
            "wdtn": "http://www.wikidata.org/prop/direct-normalized/",
            "wdv": "http://www.wikidata.org/value/",
            "rdfs": "http://www.w3.org/2000/01/rdf-schema#",
            "wikibase": "http://wikiba.se/ontology#",
            "rdfs": "http://www.w3.org/2000/01/rdf-schema#",
        }
        ns2prefix = {v: k for k, v in prefix2ns.items()}
        assert len(ns2prefix) == len(prefix2ns), "Duplicated namespaces"
        prefix_index = PrefixIndex.create(ns2prefix)

        return WikidataNamespace(prefix2ns, ns2prefix, prefix_index)

    def __init__(
        self,
        prefix2ns: dict[str, str],
        ns2prefix: dict[str, str],
        prefix_index: PrefixIndex,
    ):
        super().__init__(prefix2ns, ns2prefix, prefix_index)
        self.entity_prefix = self.ns2prefix["http://www.wikidata.org/entity/"]
        self.property_prefix = self.ns2prefix["http://www.wikidata.org/prop/"]

    ###############################################################################
    # URI testing
    ###############################################################################

    @classmethod
    def is_abs_uri_statement(cls, uri: str):
        return uri == WikidataNamespace.STATEMENT_URI

    @classmethod
    def is_abs_uri_dummy_class(cls, uri: str):
        return uri == WikidataNamespace.DUMMY_CLASS_FOR_INVERSION_URI

    @classmethod
    def is_abs_uri_property(cls, uri: str):
        return uri.startswith(f"http://www.wikidata.org/prop/P") or uri.startswith(
            f"http://www.wikidata.org/entity/P"
        )

    @classmethod
    def is_abs_uri_qnode(cls, uri: str):
        return uri.startswith("http://www.wikidata.org/entity/Q")

    @classmethod
    def is_abs_uri_lexeme(cls, uri: str):
        return uri.startswith("http://www.wikidata.org/entity/L")

    @classmethod
    def is_abs_uri_entity(cls, uri: str):
        return uri.startswith("http://www.wikidata.org/entity/")

    ###############################################################################
    # Converting between URI and ID
    ###############################################################################

    @classmethod
    def is_valid_id(cls, id: str):
        return (id[0] == "Q" or id[0] == "P" or id[0] == "L") and id[1:].isdigit()

    @classmethod
    def get_entity_id(cls, uri: str):
        if uri.startswith("http://www.wikidata.org/entity/"):
            return uri[31:]
        if uri.startswith("http://www.wikidata.org/wiki/"):
            return uri[29:]
        if uri.startswith("https://www.wikidata.org/wiki/"):
            return uri[30:]
        raise OutOfNamespace(f"{uri} is not in wikidata entity namespace")

    @classmethod
    def get_prop_id(cls, uri: str):
        if uri.startswith("http://www.wikidata.org/prop/"):
            # so this we can handle /prop/ and /prop/direct/
            return uri[uri.rfind("/") + 1 :]
        if uri.startswith("http://www.wikidata.org/entity/P"):
            return uri[31:]
        if uri.startswith("http://www.wikidata.org/wiki/Property:"):
            return uri[38:]
        if uri.startswith("https://www.wikidata.org/wiki/Property:"):
            return uri[39:]
        raise OutOfNamespace(f"{uri} is not in wikidata property namespace")

    @classmethod
    def get_entity_abs_uri(cls, iid: str):
        assert cls.is_valid_id(iid), iid
        return f"http://www.wikidata.org/entity/{iid}"

    @classmethod
    def get_prop_abs_uri(cls, pid: str):
        assert cls.is_valid_id(pid), pid
        return f"http://www.wikidata.org/prop/{pid}"

    def get_entity_rel_uri(self, iid: str):
        assert self.is_valid_id(iid), iid
        return f"{self.entity_prefix}:{iid}"

    def get_prop_rel_uri(self, pid: str):
        assert self.is_valid_id(pid), pid
        return f"{self.property_prefix}:{pid}"



from .column import Column
from .table import ColumnBasedTable
from .context import Context
from .link import EntityId, Link, WIKIDATA, WIKIDATA_NIL_ENTITY

__all__ = [
    "Column",
    "ColumnBasedTable",
    "Context",
    "EntityId",
    "Link",
    "WIKIDATA",
    "WIKIDATA_NIL_ENTITY",
]



from __future__ import annotations

from dataclasses import dataclass
from typing import Optional, Protocol, Sequence, cast


class PrecisionRecallF1Protocol(Protocol):
    @property
    def precision(self) -> float:
        ...

    @property
    def recall(self) -> float:
        ...

    @property
    def f1(self) -> float:
        ...


@dataclass
class PrecisionRecallF1:
    precision: float
    recall: float
    f1: float

    @staticmethod
    def avg(lst: Sequence[PrecisionRecallF1Protocol]) -> PrecisionRecallF1:
        n = len(lst)
        if n == 0:
            raise ValueError(
                "Cannot compute average precision-recall-f1 of an empty list"
            )

        precision = sum([x.precision for x in lst]) / n
        recall = sum([x.recall for x in lst]) / n
        f1 = sum([x.f1 for x in lst]) / n

        return PrecisionRecallF1(precision, recall, f1)

    def iter_prf(self):
        return [
            ("precision", self.precision),
            ("recall", self.recall),
            ("f1", self.f1),
        ]



from __future__ import annotations
from dataclasses import dataclass, field
from typing import List, Optional
from rsoup.core import ContentHierarchy
from sm.inputs.link import EntityId


@dataclass
class Context:
    page_title: Optional[str] = None
    page_url: Optional[str] = None
    page_entities: List[EntityId] = field(default_factory=list)
    content_hierarchy: List[ContentHierarchy] = field(default_factory=list)

    def to_dict(self):
        return {
            "version": 2,
            "page_title": self.page_title,
            "page_url": self.page_url,
            "page_entities": [e.to_dict() for e in self.page_entities],
            "content_hierarchy": [c.to_dict() for c in self.content_hierarchy],
        }

    @staticmethod
    def from_dict(obj: dict):
        version = obj.get("version")
        if version is None:
            return Context(
                page_title=obj.get("page_title"),
                page_url=obj.get("page_url"),
                page_entities=[EntityId(r, "wikidata")]
                if (r := obj.get("page_entity_id")) is not None
                else [],
                content_hierarchy=[
                    ContentHierarchy.from_dict(c)
                    for c in obj.get("content_hierarchy", [])
                ],
            )
        if version == 2:
            return Context(
                page_title=obj.get("page_title"),
                page_url=obj.get("page_url"),
                page_entities=[EntityId.from_dict(o) for o in obj["page_entities"]],
                content_hierarchy=[
                    ContentHierarchy.from_dict(c)
                    for c in obj.get("content_hierarchy", [])
                ],
            )
        raise ValueError(f"Unknown version: {version}")

    def __getstate__(self):
        return {
            "page_title": self.page_title,
            "page_url": self.page_url,
            "page_entities": self.page_entities,
            "content_hierarchy": [c.to_dict() for c in self.content_hierarchy],
        }

    def __setstate__(self, state):
        self.page_title = state["page_title"]
        self.page_url = state["page_url"]
        self.page_entities = state["page_entities"]
        self.content_hierarchy = [
            ContentHierarchy.from_dict(c) for c in state.get("content_hierarchy", [])
        ]
